// requires bible.js

"use strict";

(function (obj) {
  const bible = obj.bible;
  bible.BookNames = bible.books.map(function (book) {
    return book[0];
  });
  const numSort = (a, b) => a - b;
  const lastOtVerseIdx = 23144;

  // Credit: http://eddmann.com/posts/cartesian-product-in-javascript/
  const flatten = (arr) => [].concat.apply([], arr);
  const product = (sets) =>
    sets.reduce((acc, set) =>
      flatten(acc.map(x => set.map(y => [...x, y]))), [
        []
      ]);

  let searchExp = {};

  function buildBins(verse, verseCount) {
    let bibleBin = searchExp.bins[0];
    bibleBin.occurrences += verseCount;
    bibleBin.verses += 1;

    let testament = verse > lastOtVerseIdx ? 'New Testament' : 'Old Testament';
    let testamentBin = bibleBin.items.find(x => x.name === testament);
    testamentBin.occurrences += verseCount;
    testamentBin.verses += 1;

    let ref = bible.refs[verse];
    let matches = /(.*) (\d+):\d+/.exec(ref);
    let book = matches[1];
    let chapterNum = matches[2];
    let bookIdx = bible.BookNames.indexOf(book);
    let chapterIdx = chapterNum - 1;
    let chapter = 'Chapter ' + chapterNum;

    let bookBin = testamentBin.items.find(x => x.name === book);
    if (!bookBin) {
      testamentBin.items.push({
        name: book,
        citationName: book,
        start: bible.books[bookIdx][2],
        end: bible.books[bookIdx][3],
        occurrences: 0,
        verses: 0,
        items: []
      });
      bookBin = testamentBin.items[testamentBin.items.length - 1];
    }
    bookBin.occurrences += verseCount;
    bookBin.verses += 1;

    let chapterBin = bookBin.items.find(x => x.name === chapter);
    if (!chapterBin) {
      bookBin.items.push({
        name: chapter,
        citationName: `${book} ${chapterNum}`,
        start: bible.books[bookIdx][4][chapterIdx][0],
        end: bible.books[bookIdx][4][chapterIdx][1],
        occurrences: 0,
        verses: 0
      });
      chapterBin = bookBin.items[bookBin.items.length - 1];
    }
    chapterBin.occurrences += verseCount;
    chapterBin.verses += 1;
  }

  function buildRegExp(term, flags) {
    let regexStr = term.replace(/\*/g, '[\\w\']*');
    regexStr = term.endsWith('*') ?
      `\\b(${regexStr})` :
      `\\b(${regexStr})( |$)`;
    return new RegExp(regexStr, flags);
  }

  function buildSearchCombinations() {
    searchExp.combinations = product(searchExp.patterns);
  }

  function buildSearchExp(searchStr) {
    searchExp.state = 'ERROR';
    searchExp.searchStr = searchStr;
    if (searchStr === "") {
      searchExp.type = 'EMPTY';
      return searchExp;
    }
    searchExp.type = 'INVALID';
    searchExp.flags = searchStr.startsWith('@') ? 'g' : 'gi';
    searchExp.searchTerms = searchExp.searchStr
      .replace('@', '')
      .trim()
      .replace(/ {2,}/g, ' ')
      .replace(/ *,+ */g, ',');
    if (searchExp.searchTerms.match(/[^a-z ,'*-]/i)) {
      return searchExp;
    }
    if (
      searchExp.searchTerms.includes(' ') &&
      searchExp.searchTerms.includes(',')
    ) {
      return searchExp;
    }
    searchExp.terms = searchExp.searchTerms.replace(/-/g, '').split(/[ ,]/);
    if (
      searchExp.searchStr.includes(',') ||
      searchExp.terms.length === 1
    ) {
      searchExp.type = 'WORD';
    } else {
      searchExp.type = 'PHRASE';
    }
    return searchExp;
  }

  function buildSearchIntersects() {
    let verses = new Set();
    searchExp.sets.forEach(function (set) {
      let intersect = intersectAll(set);
      [...intersect].forEach(function (verse) {
        verses.add(verse);
      });
    });
    searchExp.intersects = [...verses].sort(numSort);
  }

  function buildSearchPatterns() {
    searchExp.bibleWords = 'OK';
    searchExp.patterns = [];
    let missingTerms = [];
    searchExp.terms.forEach(function (term) {
      let regExp = buildRegExp(term, 'gi');
      let words = getbibleWords(regExp);
      if (words.length > 0) {
        searchExp.patterns.push(words);
      } else {
        missingTerms.push(term);
      }
    });
    if (missingTerms.length > 0) {
      searchExp.bibleWords = `'${missingTerms.join(', ')}' not found`;
    }
  }

  function buildSearchPhraseVerses() {
    let allVerses = [...searchExp.intersects].sort(numSort);
    allVerses.forEach(function (verse) {
      let text = bible.verses[verse].replace(/[!();:,.?-]/g, '');
      let regExp = buildRegExp(searchExp.searchTerms, searchExp.flags);
      let verseCount = (text.match(regExp) || []).length;
      if (verseCount > 0) {
        searchExp.verses.push(verse);
        buildBins(verse, verseCount);
      }
    });
  }

  function buildSearchSets() {
    searchExp.sets = [];
    searchExp.combinations.forEach(function (combination) {
      let comboSets = [];
      combination.forEach(function (word) {
        comboSets.push(new Set(bible.wordVerses[word]));
      });
      searchExp.sets.push(comboSets);
    });
  }

  function buildSearchWordVerses() {
    let allVerses = [...searchExp.intersects].sort(numSort);
    allVerses.forEach(function (verse) {
      let text = bible.verses[verse].replace(/[!();:,.?-]/g, '');
      let verseCount = 0;
      searchExp.terms.every(function (term) {
        let regExp = buildRegExp(term, searchExp.flags);
        let hits = (text.match(regExp) || []).length;
        if (hits === 0) {
          verseCount = 0;
          return false;
        } else {
          verseCount += hits;
          return true;
        }
      });
      if (verseCount > 0) {
        searchExp.verses.push(verse);
        buildBins(verse, verseCount);
      }
    });
  }

  function buildSearchVerses() {
    searchExp.occurrences = 0;
    searchExp.verses = [];
    initializeBins();
    switch (searchExp.type) {
      case 'PHRASE':
        buildSearchPhraseVerses(searchExp);
        break;
      case 'WORD':
        buildSearchWordVerses(searchExp);
        break;
    }
  }

  function findAllMatches(str, regEx) {
    let result;
    let matches = [];
    while ((result = regEx.exec(str)) !== null) {
      matches.push(result[1]);
    }
    return matches.length === 0 ? undefined : matches;
  }

  function getbibleWords(regExp) {
    let bibleWords = [];
    let words = findAllMatches(bible.wordList, regExp);
    if (words) {
      bibleWords = bibleWords.concat(words);
    }
    return bibleWords;
  }

  function initializeBins() {
    searchExp.bins = [{
      name: 'Bible',
      citationName: 'the Bible',
      start: 0,
      end: bible.verses.length - 1,
      occurrences: 0,
      verses: 0,
      items: [{
        name: 'Old Testament',
        citationName: 'the Old Testament',
        start: 0,
        end: lastOtVerseIdx,
        occurrences: 0,
        verses: 0,
        items: []
      }, {
        name: 'New Testament',
        citationName: 'the New Testament',
        start: lastOtVerseIdx + 1,
        end: bible.verses.length - 1,
        occurrences: 0,
        verses: 0,
        items: []
      }]
    }];
  }

  function intersectAll(...sets) {
    let numOfSets = sets.length;
    if (numOfSets == 0) {
      return undefined;
    }
    if (Array.isArray(sets[0])) {
      sets = [...sets[0]];
      numOfSets = sets.length;
    }
    if (numOfSets < 2) {
      return sets[0];
    }
    let intersect = intersection(sets[0], sets[1]);
    if (numOfSets == 2) {
      return intersect;
    }
    for (let i = 2; i < numOfSets; i++) {
      intersect = intersection(intersect, sets[i]);
    }
    return intersect;
  }

  function intersection(set1, set2) {
    return new Set([...set1].filter(x => set2.has(x)));
  }

  obj.performSearch = function (searchStr) {
    searchExp = buildSearchExp(searchStr);
    if (
      searchExp.type === 'EMPTY' ||
      searchExp.type === 'INVALID'
    ) {
      return searchExp;
    }
    buildSearchPatterns();
    if (searchExp.bibleWords !== 'OK') {
      return searchExp;
    }
    searchExp.state = 'OK';
    buildSearchCombinations();
    buildSearchSets();
    buildSearchIntersects();
    buildSearchVerses();
    return searchExp;
  };
})(window.KJV);
