// requires search.js

"use strict";

(function (obj) {
  const bible = obj.bible;
  const performSearch = obj.performSearch;
  const numSortAscending = (a, b) => a - b;
  const numSortDescending = (a, b) => b - a;

  const binCopy = document.querySelector('.bin-copy');
  const bookmarksEmpty = document.querySelector('.bookmarks-empty');
  const bookmarksList = document.querySelector('.bookmarks-list');
  const btnBookmarkDelete = document.querySelector('.btn-bookmark-delete');
  const btnBookmarkDown = document.querySelector('.btn-bookmark-down');
  const btnBookmarkSelect = document.querySelector('.btn-bookmark-select');
  const btnBookmarkUp = document.querySelector('.btn-bookmark-up');
  const btnBookmarks = document.querySelector('.btn-bookmarks');
  const btnBookmarksBack = document.querySelector('.btn-bookmarks-back');
  const btnBookmarksSortAscending =
    document.querySelector('.btn-bookmarks-sort-ascending');
  const btnBookmarksSortDescending =
    document.querySelector('.btn-bookmarks-sort-descending');
  const btnColumnOne = document.querySelector('.btn-column-one');
  const btnColumnThree = document.querySelector('.btn-column-three');
  const btnColumnTwo = document.querySelector('.btn-column-two');
  const btnContents = document.querySelector('.btn-contents');
  const btnContentsBack = document.querySelector('.btn-contents-back');
  const btnFontName = document.querySelector('.btn-font-name');
  const btnFontNext = document.querySelector('.btn-font-next');
  const btnFontPrev = document.querySelector('.btn-font-prev');
  const btnHelp = document.querySelector('.btn-help');
  const btnHelpAbout = document.querySelector('.btn-help-about');
  const btnHelpBack = document.querySelector('.btn-help-back');
  const btnHelpRead = document.querySelector('.btn-help-read');
  const btnHelpSearch = document.querySelector('.btn-help-search');
  const btnHelpSettings = document.querySelector('.btn-help-settings');
  const btnNt = document.querySelector('.btn-nt');
  const btnOt = document.querySelector('.btn-ot');
  const btnSearch = document.querySelector('.btn-search');
  const btnSearchBack = document.querySelector('.btn-search-back');
  const btnSearchGo = document.querySelector('.btn-search-go');
  const chapterList = document.querySelector('.chapter-list');
  const detailsBanner = document.querySelector('.details-banner');
  const detailsList = document.querySelector('.details-list');
  const fontSample = document.querySelector('.font-sample');
  const helpAbout = document.querySelector('.help-about');
  const helpRead = document.querySelector('.help-read');
  const helpSearch = document.querySelector('.help-search');
  const helpSettings = document.querySelector('.help-settings');
  const html = document.documentElement;
  const navBookmarks = document.querySelector('.nav-bookmarks');
  const navContents = document.querySelector('.nav-contents');
  const navFont = document.querySelector('.nav-font');
  const navHelp = document.querySelector('.nav-help');
  const navRead = document.querySelector('.nav-read');
  const navSearch = document.querySelector('.nav-search');
  const ntList = document.querySelector('.nt-list');
  const otList = document.querySelector('.ot-list');
  const paneBookmarks = document.querySelector('.pane-bookmarks');
  const paneContents = document.querySelector('.pane-contents');
  const paneHelp = document.querySelector('.pane-help');
  const paneRead = document.querySelector('.pane-read');
  const paneSearch = document.querySelector('.pane-search');
  const readStats = document.querySelector('.read-stats');
  const readVerses = document.querySelector('.read-verses');
  const resultsBanner = document.querySelector('.results-banner');
  const resultsList = document.querySelector('.results-list');
  const scrollBookmarks = document.querySelector('.scroll-bookmarks');
  const scrollContents = document.querySelector('.scroll-contents');
  const scrollHelp = document.querySelector('.scroll-help');
  const scrollRead = document.querySelector('.scroll-read');
  const scrollSearch = document.querySelector('.scroll-search');
  const searchDetails = document.querySelector('.search-details');
  const searchInput = document.querySelector('.search-input');
  const searchResults = document.querySelector('.search-results');
  const selectorBookmarks = document.querySelector('.selector-bookmarks');
  const selectorContents = document.querySelector('.selector-contents');
  const selectorRead = document.querySelector('.selector-read');
  const selectorSearch = document.querySelector('.selector-search');

  const fontClasses = [
    'font-slabo',
    'font-roboto-slab',
    'font-merriweather',
    'font-gabriela',
    'font-almendra',
    'font-roboto',
    'font-open-sans',
    'font-lato',
    'font-inconsolata',
    'font-source-code-pro',
    'font-indie-flower',
    'font-give-you-glory'
  ];
  const fontMaxIdx = fontClasses.length - 1;
  const fontNames = fontClasses
    .map(function (x) {
      return x
        .substring(5)
        .split('-')
        .map(function (y) {
          return y.charAt(0).toUpperCase() + y.slice(1);
        })
        .join(' ');
    });

  const sidebarBookmarks = 1;
  const sidebarContents = 0;
  const sidebarHelp = 3;
  const sidebarSearch = 2;
  const sidebars = [sidebarContents, sidebarBookmarks, sidebarSearch,
    sidebarHelp
  ];

  const columnBtns = [btnColumnOne, btnColumnTwo, btnColumnThree];
  const helpBtns = [btnHelpAbout, btnHelpRead, btnHelpSearch, btnHelpSettings];
  const helpPages = [helpAbout, helpRead, helpSearch, helpSettings];
  const sidebarBtns = [btnContents, btnBookmarks, btnSearch, btnHelp];
  const sidebarPanes = [paneContents, paneBookmarks, paneSearch, paneHelp];

  const firstNtBookIdx = 39;
  const firstNtChapterIdx = 929;
  const firstNtVerseIdx = 23145;
  const maxChapters = 150;

  let activeBook = null;
  let activeBookmark = null;
  let activeChapter = null;
  let activeColumnIdx = null;
  let activeFontIdx = null;
  let activeHelpIdx = null;
  let activeSidebarIdx = null;
  let activeTestament = null;
  let bibleStats = {};
  let bookmarks;
  let currentPanes = null;
  let fragment;
  let numPanes = null;
  let searchExp = null;
  let searchFilter = {};
  let searchStr = null;
  let selectedBookChapter = {};
  let selectingBookmarks = false;

  function activate(element) {
    element.classList.add('active');
  }

  function activateBibleDetail() {
    let btn = detailsList.querySelector('li>span+button');
    if (btn) {
      btn.classList.add('active-detail');
    }
  }

  function addBookmark(verseIdx) {
    if (bookmarks.indexOf(verseIdx) !== -1) {
      return;
    }
    bookmarks = [verseIdx, ...bookmarks];
    saveBookmarks();
  }

  function addListeners() {
    navBookmarks.addEventListener('click', navBookmarksClick);
    navContents.addEventListener('click', navContentsClick);
    navFont.addEventListener('click', navFontClick);
    navHelp.addEventListener('click', navHelpClick);
    navRead.addEventListener('click', navReadClick);
    navSearch.addEventListener('click', navSearchClick);
    searchInput.addEventListener('keydown', searchInputKeyDown);
    selectorBookmarks.addEventListener('click', selectorBookmarksClick);
    selectorContents.addEventListener('click', selectorContentsClick);
    selectorRead.addEventListener('click', selectorReadClick);
    selectorSearch.addEventListener('click', selectorSearchClick);
    window.addEventListener('resize', updatePanes);
  }

  function addReadVerse(verseIdx, verseNum) {
    let verse = document.createElement('li');
    let verseNumSpan = buildVerseNumSpan(verseNum);
    let acrostic = buildAcrosticSpan(verseIdx);
    let text = document.createTextNode(bible.verses[verseIdx]);
    verse.appendChild(verseNumSpan);
    if (acrostic) {
      verse.appendChild(acrostic);
    }
    verse.appendChild(text);
    verse.setAttribute('data-verse', verseIdx);
    verse.classList.add('read-verse');
    if (bookmarks.indexOf(verseIdx) !== -1) {
      verse.classList.add('bookmark');
    }
    fragment.appendChild(verse);
  }

  function addSearchVerse(verseIdx) {
    let verse = document.createElement('li');
    let acrostic = buildAcrosticSpan(verseIdx);
    let ref = buildRefSpan(verseIdx);
    let text = document.createTextNode(bible.verses[verseIdx]);
    verse.appendChild(ref);
    if (acrostic) {
      verse.appendChild(acrostic);
    }
    verse.appendChild(text);
    fragment.appendChild(verse);
  }

  function btnBookClick(event) {
    let target = event.target;
    if (target === activeBook) {
      chapterList.classList.toggle('hide');
      return;
    }
    deactivate(activeBook);
    activeBook = target;
    let bookIndex = getDataBook(activeBook);
    let chapterCount = bible.books[bookIndex][1];
    resetChapterList(chapterCount);
    resetActiveChapter();
    activate(activeBook);
    activeBook.parentNode.insertBefore(chapterList, activeBook.nextSibling);
    show(chapterList);
    centerScrollElement(scrollContents, activeBook);
    if (chapterCount === 1) {
      let btnChapterOne = chapterList.querySelector('.btn-chapter');
      btnChapterOne.click();
    }
  }

  function btnBookmarkDeleteClick() {
    if (!activeBookmark) {
      return;
    }
    let verseIdx = getDataVerse(activeBookmark);
    let bookmarksIdx = bookmarks.indexOf(verseIdx);
    bookmarks.splice(bookmarksIdx, 1);
    updateBookmarkList();
    removeBookmarkHighlight(verseIdx);
    if (bookmarks.length === 0) {
      activeBookmark = null;
      return;
    }
    let nextBookmarksIdx = Math.min(bookmarks.length - 1, bookmarksIdx);
    activeBookmark = bookmarksList.querySelector(
      `button[data-verse="${bookmarks[nextBookmarksIdx]}"]`);
    activate(activeBookmark);
  }

  function btnBookmarkDownClick(target) {
    if (!activeBookmark) {
      return;
    }
    let verseIdx = getDataVerse(activeBookmark);
    let bookmarksIdx = bookmarks.indexOf(verseIdx);
    if (bookmarksIdx === (bookmarks.length - 1) || bookmarksIdx === -1) {
      return;
    }
    reorderBookmarks(bookmarksIdx, bookmarksIdx + 1);
    saveBookmarks();
    updateBookmarkList();
    activeBookmark = bookmarksList.querySelector(
      `button[data-verse="${verseIdx}"]`);
    activate(activeBookmark);
  }

  function btnBookmarksSelectClick(target) {
    if (bookmarks.length === 0) {
      return;
    }
    if (selectingBookmarks) {
      if (activeBookmark) {
        deactivate(activeBookmark);
        activeBookmark = null;
      }
      deactivate(btnBookmarkSelect);
      show(btnBookmarksSortAscending);
      show(btnBookmarksSortDescending);
      hide(btnBookmarkUp);
      hide(btnBookmarkDown);
      hide(btnBookmarkDelete);
      selectingBookmarks = false;
      return;
    }
    activate(btnBookmarkSelect);
    hide(btnBookmarksSortAscending);
    hide(btnBookmarksSortDescending);
    show(btnBookmarkUp);
    show(btnBookmarkDown);
    show(btnBookmarkDelete);
    selectingBookmarks = true;
  }

  function btnBookmarkUpClick() {
    if (!activeBookmark) {
      return;
    }
    let verseIdx = getDataVerse(activeBookmark);
    let bookmarksIdx = bookmarks.indexOf(verseIdx);
    if (bookmarksIdx === 0 || bookmarksIdx === -1) {
      return;
    }
    reorderBookmarks(bookmarksIdx, bookmarksIdx - 1);
    updateBookmarkList();
    saveBookmarks();
    activeBookmark = bookmarksList.querySelector(
      `button[data-verse="${verseIdx}"]`);
    activate(activeBookmark);
  }

  function btnChapterClick(event) {
    let target = event.target;
    if (target === activeChapter) {
      if (numPanes === 1) {
        hide(paneContents);
        show(paneRead);
      }
      return;
    }
    if (activeChapter) {
      deactivate(activeChapter);
    }
    activeChapter = target;
    let chapterIdx = getDataChapter(target);
    let bookIdx = getDataBook(activeBook);
    setSelectedBookChapter(bookIdx, chapterIdx);
    saveSelectedBookChapter();
    activate(activeChapter);
    if (numPanes === 1) {
      hide(paneContents);
      show(paneRead);
    }
    updatePaneRead();
    scrollRead.scrollTop = 0;
  }

  function btnSearchGoClick(event) {
    searchStr = searchInput.value;
    saveSearchStr();
    updateSearchPane();
  }

  function buildAcrosticSpan(verseIdx) {
    let acrostic = bible.acrostics[verseIdx];
    if (!acrostic) {
      return undefined;
    }
    let acrosticSpan = document.createElement('span');
    acrosticSpan.classList.add('acrostic');
    acrosticSpan.textContent = acrostic;
    return acrosticSpan;
  }

  function buildAcrosticText(verseIdx) {
    let acrostic = bible.acrostics[verseIdx];
    if (!acrostic) {
      return undefined;
    }
    return acrostic;
  }

  function buildBibleSearchFilter() {
    if (searchExp.state !== 'OK') {
      searchFilter = {};
      return;
    }
    let bible = searchExp.bins[0];
    searchFilter['citation'] =
      `${bible.occurrences} / ${bible.verses} in the Bible`;
    searchFilter['verses'] = searchExp.verses;
  }

  function buildBibleStats() {
    let bibleChapterNum = 0;
    bible.books.forEach(function (book, bookIdx) {
      let bookName = book[0];
      let chapters = book[4];
      chapters.forEach(function (chapter, chapterIdx) {
        bibleChapterNum += 1;
        let statName = `${bookName} ${chapterIdx + 1}`;
        let stats = {};
        stats['bibleBook'] = bookIdx + 1;
        stats['bibleChapter'] = bibleChapterNum;
        stats['bibleVerse'] = chapter[0] + 1;
        if (bookIdx >= firstNtBookIdx) {
          stats['ntBook'] = bookIdx - firstNtBookIdx + 1;
          stats['ntChapter'] = bibleChapterNum - firstNtChapterIdx;
          stats['ntVerse'] = chapter[0] - firstNtVerseIdx + 1;
        } else {
          stats['ntBook'] = 0;
          stats['ntChapter'] = 0;
          stats['ntVerse'] = 0;
        }
        bibleStats[statName] = stats;
      });
    });
  }

  function buildBtnBook(bookIndex) {
    let btn = document.createElement('button');
    btn.classList.add('btn', 'btn-book');
    btn.setAttribute('data-book', bookIndex);
    btn.textContent = bible.books[bookIndex][0];
    return btn;
  }

  function buildBtnBookmark(verseIdx) {
    let btn = document.createElement('button');
    btn.classList.add('btn', 'btn-bookmark');
    btn.setAttribute('data-verse', verseIdx);
    btn.textContent = bible.refs[verseIdx];
    return btn;
  }

  function buildBtnChapter(chapterIndex) {
    let btn = document.createElement('button');
    btn.classList.add('btn', 'btn-chapter');
    btn.setAttribute('data-chapter', chapterIndex);
    btn.textContent = (chapterIndex + 1).toString();
    return btn;
  }

  function buildChapterList() {
    let btn;
    fragment = document.createDocumentFragment();
    for (let chapterIndex = 0; chapterIndex < maxChapters; chapterIndex++) {
      btn = buildBtnChapter(chapterIndex);
      fragment.appendChild(btn);
    }
    chapterList.appendChild(fragment);
  }

  function buildClipboardText(verseIdx) {
    let text = buildRefText(verseIdx);
    let acrostic = buildAcrosticText(verseIdx);
    if (acrostic) {
      text += ' ' + acrostic;
    }
    return text + ' ' + buildVerseText(verseIdx);
  }

  function buildDetailsItem(bin) {
    let item = document.createElement('li');
    item.setAttribute('data-citation',
      `${bin.occurrences} / ${bin.verses} in ${bin.citationName}`);
    item.setAttribute('data-start', bin.start);
    item.setAttribute('data-end', bin.end);
    let span = document.createElement('span');
    span.textContent =
      `${bin.name}: ${bin.occurrences} / ${bin.verses}`;
    item.appendChild(span);
    let btn = document.createElement('button');
    btn.classList.add('btn-filter');
    item.appendChild(btn);
    return item;
  }

  function buildOtList() {
    let btn;
    fragment = document.createDocumentFragment();
    for (let bookIndex = 0; bookIndex < firstNtBookIdx; bookIndex++) {
      btn = buildBtnBook(bookIndex);
      fragment.appendChild(btn);
    }
    otList.appendChild(fragment);
  }

  function buildNtList() {
    let btn;
    let lastIdx = bible.books.length;
    fragment = document.createDocumentFragment();
    for (let bookIndex = firstNtBookIdx; bookIndex < lastIdx; bookIndex++) {
      btn = buildBtnBook(bookIndex);
      fragment.appendChild(btn);
    }
    ntList.appendChild(fragment);
  }

  function buildRefSpan(verseIdx) {
    let refSpan = document.createElement('span');
    refSpan.classList.add('ref');
    refSpan.textContent = bible.refs[verseIdx];
    return refSpan;
  }

  function buildRefText(verseIdx) {
    return bible.refs[verseIdx];
  }

  function buildSearchDetailsList(bins) {
    let list = document.createElement('ul');
    bins.forEach(function (bin) {
      let item = buildDetailsItem(bin);
      if (bin.items) {
        let children = buildSearchDetailsList(bin.items);
        if (bin.name === 'Bible') {
          item.classList.add('fold');
        } else {
          item.classList.add('unfold');
        }
        item.appendChild(children);
      } else {
        list.classList.add('chapter-list');
      }
      list.appendChild(item);
    });
    return list;
  }

  function buildSearchFilter(element) {
    searchFilter['citation'] = getDataCitation(element);
    searchFilter['verses'] = filteredSearchVerses(
      getDataStart(element), getDataEnd(element)
    );
  }

  function buildVerseNumSpan(verseNum) {
    let verseNumSpan = document.createElement('span');
    verseNumSpan.classList.add('verse-num');
    verseNumSpan.textContent = verseNum;
    return verseNumSpan;
  }

  function buildVerseText(verseIdx) {
    return bible.verses[verseIdx];
  }

  function centerScrollElement(scrollElement, element) {
    let y = element.offsetTop -
      (scrollElement.clientHeight - element.clientHeight) / 2;
    scrollElement.scrollTop = y;
  }

  function copyToClipboard(verseIdx) {
    binCopy.value = buildClipboardText(verseIdx);
    if (navigator.userAgent.match(/ipad|ipod|iphone/i)) {
      binCopy.setSelectionRange(0, 999999);
    } else {
      binCopy.select();
    }
    document.execCommand('copy');
  }

  function deactivate(element) {
    element.classList.remove('active');
  }

  function deactivateDetail() {
    let btn = detailsList.querySelector('button.active-detail');
    if (btn) {
      btn.classList.remove('active-detail');
    }
  }

  function decodeRef(ref) {
    let parts = ref.match(/([\w ]+) (\d+):(\d+)/);
    let book = bible.books.findIndex(a => a[0] === parts[1]);
    let chapter = parseInt(parts[2]) - 1;
    let verse = parseInt(parts[3]) - 1;
    return {
      book: book,
      chapter: chapter,
      verse: verse
    };
  }

  function decreasePanes() {
    if (numPanes === 1) {
      hideActiveSidebar();
      setActiveSidebarIdx(null);
      updateColumns(1);
      show(paneRead);
      return;
    }
    if (numPanes === 3 && activeColumnIdx > 1) {
      updateColumns(2);
      return;
    }
    if (numPanes === 2 && activeColumnIdx > 0) {
      updateColumns(1);
      return;
    }
  }

  function detailsBannerClick(target) {
    hide(searchDetails);
    show(searchResults);
  }

  function detailsListClick(target) {
    if (target.tagName === 'BUTTON') {
      filterBtnClick(target);
      return;
    }
    if (target.tagName === 'SPAN') {
      target = target.closest('li');
    }
    if (target.tagName === 'LI') {
      let children = target.querySelector('ul');
      if (children) {
        if (target.classList.contains('unfold')) {
          target.classList.remove('unfold');
          target.classList.add('fold');
        } else {
          target.classList.remove('fold');
          target.classList.add('unfold');
        }
      }
    }
  }

  function filterBtnClick(target) {
    target = target.closest('li');
    let btn = target.querySelector('button');
    deactivateDetail();
    btn.classList.add('active-detail');
    buildSearchFilter(target);
    updateSearchResults();
    hide(searchDetails);
    show(searchResults);
    scrollSearch.scrollTop = 0;
  }

  function filteredSearchVerses(start, end) {
    return searchExp.verses.filter(function (verse) {
      return verse >= start && verse <= end;
    });
  }

  function getActiveColumnIdx() {
    let value = localStorage.getItem('activeColumnIdx');
    if (!value) {
      activeColumnIdx = 0;
      saveActiveColumnIdx();
      return;
    }
    activeColumnIdx = JSON.parse(value);
  }

  function getActiveFontIdx() {
    let value = localStorage.getItem('activeFontIdx');
    if (!value) {
      activeFontIdx = fontClasses.indexOf('font-roboto');
      setActiveFontIdx();
      saveActiveFontIdx();
      return;
    }
    activeFontIdx = JSON.parse(value);
  }

  function getActiveHelpIdx() {
    let value = localStorage.getItem('activeHelpIdx');
    if (!value) {
      activeHelpIdx = 0;
      saveActiveHelpIdx();
      return;
    }
    activeHelpIdx = JSON.parse(value);
  }

  function getActiveSidebarIdx() {
    let value = localStorage.getItem('activeSidebarIdx');
    if (!value) {
      setActiveSidebarIdx(null);
      return;
    }
    activeSidebarIdx = JSON.parse(value);
  }

  function getBookmarks() {
    let value = localStorage.getItem('bookmarks');
    if (!value) {
      bookmarks = [];
      return;
    }
    bookmarks = JSON.parse(value);
  }

  function getDataBook(element) {
    return parseInt(element.getAttribute('data-book'));
  }

  function getDataChapter(element) {
    return parseInt(element.getAttribute('data-chapter'));
  }

  function getDataCitation(element) {
    return element.getAttribute('data-citation');
  }

  function getDataEnd(element) {
    return parseInt(element.getAttribute('data-end'));
  }

  function getDataStart(element) {
    return parseInt(element.getAttribute('data-start'));
  }

  function getDataVerse(element) {
    return parseInt(element.getAttribute('data-verse'));
  }

  function getNextFontIdx() {
    if (activeFontIdx === fontMaxIdx) {
      return 0;
    }
    return activeFontIdx + 1;
  }

  function getNumPanes() {
    numPanes = Math.min(Math.floor(window.innerWidth / 320), 4);
  }

  function getPrevFontIdx() {
    if (activeFontIdx === 0) {
      return fontMaxIdx;
    }
    return activeFontIdx - 1;
  }

  function getSearchStr() {
    let value = localStorage.getItem('searchStr');
    if (!value) {
      searchStr = "";
      saveSearchStr();
      return;
    }
    searchStr = JSON.parse(value);
    saveSearchStr();
  }

  function getSelectedBookChapter() {
    let value = localStorage.getItem('selectedBookChapter');
    if (!value) {
      setSelectedBookChapter(0, 0);
      saveSelectedBookChapter();
      return;
    }
    selectedBookChapter = JSON.parse(value);
  }

  function goToBookmark(verseIdx) {
    let decode = decodeRef(bible.refs[verseIdx]);
    setSelectedBookChapter(decode.book, decode.chapter);
    saveSelectedBookChapter();
    updateSelectorContents();
    updatePaneRead();
    if (numPanes === 1) {
      hide(paneBookmarks);
    }
    show(paneRead);
    scrollToVerse(scrollRead, decode.verse);
  }

  function hide(element) {
    element.classList.add('hide');
  }

  function hideActiveHelp() {
    deactivate(helpBtns[activeHelpIdx]);
    hide(helpPages[activeHelpIdx]);
  }

  function highlightVerse(verseIdx) {
    let verses = readVerses.querySelectorAll('li');
    verses[verseIdx].classList.add('search-flash');
    setTimeout(function () {
      verses[verseIdx].classList.remove('search-flash');
    }, 4000);
  }

  function hideActiveSidebar() {
    hide(sidebarPanes[activeSidebarIdx]);
    deactivate(sidebarBtns[activeSidebarIdx]);
  }

  function increasePanes() {
    if (currentPanes > 1) {
      return;
    }
    if (activeSidebarIdx !== null) {
      show(paneRead);
      updateSidebar(activeSidebarIdx);
      return;
    }
    updateSidebar(sidebarContents);
    return;
  }

  function navBookmarksClick(event) {
    event.preventDefault();
    let target = event.target;
    if (target === navBookmarks) {
      return;
    }
    target = target.closest('button');
    switch (target) {
      case btnBookmarksBack:
        hide(paneBookmarks);
        show(paneRead);
        return;
        break;
      case btnBookmarksSortAscending:
        sortBookmarks(numSortAscending);
        break;
      case btnBookmarksSortDescending:
        sortBookmarks(numSortDescending);
        break;
      case btnBookmarkSelect:
        btnBookmarksSelectClick(target);
        break;
      case btnBookmarkUp:
        btnBookmarkUpClick();
        break;
      case btnBookmarkDown:
        btnBookmarkDownClick();
        break;
      case btnBookmarkDelete:
        btnBookmarkDeleteClick();
        break;
      default:
        return;
    }
    saveBookmarks();
  }

  function navContentsClick(event) {
    event.preventDefault();
    let target = event.target;
    if (target === navContents) {
      return;
    }
    target = target.closest('button');
    if (target === btnContentsBack) {
      hide(paneContents);
      show(paneRead);
      updateSelectorContents();
      return;
    }
    if (target === activeTestament) {
      return;
    }
    switch (target) {
      case btnOt:
        updateBtnOt();
        break;
      case btnNt:
        updateBtnNt();
        break;
    }
  }

  function navFontClick(event) {
    event.preventDefault();
    let target = event.target;
    if (target === navFont || target.classList.contains('btn-font-name')) {
      return;
    }
    target = target.closest('button');
    if (target.classList.contains('btn-font-next')) {
      activeFontIdx = getNextFontIdx();
    } else {
      activeFontIdx = getPrevFontIdx();
    }
    let activeFont = fontClasses[activeFontIdx];
    saveActiveFontIdx();
    setActiveFontIdx();
  }

  function navHelpClick(event) {
    event.preventDefault();
    let target = event.target;
    let activeHelpBtn = helpBtns[activeHelpIdx];
    if (target === navHelp || activeHelpBtn === target) {
      return;
    }
    target = target.closest('button');
    if (target === btnHelpBack) {
      hide(paneHelp);
      show(paneRead);
      return;
    }
    scrollHelp.scrollTop = 0;
    hideActiveHelp();
    activeHelpBtn = target;
    activeHelpIdx = helpBtns.indexOf(activeHelpBtn);
    saveActiveHelpIdx();
    showActiveHelp();
  }

  function navReadClick(event) {
    event.preventDefault();
    let target = event.target;
    if (target === navRead) {
      return;
    }
    target = target.closest('button');
    if (numPanes === 1) {
      hide(paneRead);
    }
    switch (target) {
      case btnContents:
        updateSidebar(sidebarContents);
        centerScrollElement(scrollContents, activeChapter);
        break;
      case btnBookmarks:
        updateSidebar(sidebarBookmarks);
        break;
      case btnSearch:
        updateSidebar(sidebarSearch);
        break;
      case btnHelp:
        updateSidebar(sidebarHelp);
        break;
      case btnColumnOne:
      case btnColumnTwo:
      case btnColumnThree:
        let columnCount = columnBtns.indexOf(target) + 1;
        updateColumns(columnCount);
        break;
      default:
        break;
    }
  }

  function navSearchClick(event) {
    event.preventDefault();
    let target = event.target;
    if (target === navSearch || target === searchInput) {
      return;
    }
    target = target.closest('button');
    switch (target) {
      case btnSearchBack:
        hide(paneSearch);
        show(paneRead);
        break;
      case btnSearchGo:
        btnSearchGoClick();
        break;
    }
  }

  function readVerseClick(target) {
    let text = window.getSelection().toString();
    if (text) {
      return;
    }
    let verseIdx = getDataVerse(target);
    if (target.classList.contains('bookmark')) {
      target.classList.remove('bookmark');
      removeBookmark(verseIdx);
    } else {
      target.classList.add('bookmark');
      addBookmark(verseIdx);
    }
    updateBookmarkList();
  }

  function removeAllChildren(element) {
    while (element.firstChild) {
      element.removeChild(element.firstChild);
    }
  }

  function removeBookmark(verseIdx) {
    let index = bookmarks.indexOf(verseIdx);
    if (index === -1) {
      return;
    }
    bookmarks.splice(index, 1);
    saveBookmarks();
  }

  function removeBookmarkHighlight(verseIdx) {
    let highlighted = readVerses.querySelector(
      `li[data-verse="${verseIdx}"]`);
    if (highlighted) {
      highlighted.classList.remove('bookmark');
    }
  }

  function reorderBookmarks(fromIdx, toIdx) {
    bookmarks.splice(toIdx, 0, bookmarks.splice(fromIdx, 1)[0]);
  }

  function resetActiveChapter() {
    let activeBookIdx = getDataBook(activeBook);
    if (activeBookIdx === selectedBookChapter.bookIdx) {
      activeChapter = chapterList.querySelector(
        `.btn-chapter[data-chapter="${selectedBookChapter.chapterIdx}"]`
      );
      activate(activeChapter);
    }
  }

  function resetChapterList(chapterCount) {
    let buttons = chapterList.querySelectorAll('.btn-chapter');
    buttons.forEach(function (btn, index) {
      if (index < chapterCount) {
        show(btn);
        return;
      } else {
        hide(btn);
        return;
      }
    });
    if (activeChapter) {
      deactivate(activeChapter);
      activeChapter = null;
    }
  }

  function saveActiveColumnIdx() {
    localStorage.setItem('activeColumnIdx', JSON.stringify(activeColumnIdx));
  }

  function saveActiveFontIdx() {
    localStorage.setItem('activeFontIdx', JSON.stringify(activeFontIdx));
  }

  function saveActiveHelpIdx() {
    localStorage.setItem('activeHelpIdx', JSON.stringify(activeHelpIdx));
  }

  function saveActiveSidebarIdx() {
    localStorage.setItem('activeSidebarIdx', JSON.stringify(activeSidebarIdx));
  }

  function saveBookmarks() {
    localStorage.setItem('bookmarks', JSON.stringify(bookmarks));
  }

  function saveSearchStr() {
    localStorage.setItem('searchStr', JSON.stringify(searchStr));
  }

  function saveSelectedBookChapter() {
    localStorage.setItem(
      'selectedBookChapter', JSON.stringify(selectedBookChapter));
  }

  function scrollToVerse(pane, verseIdx) {
    let verses = readVerses.querySelectorAll('li');
    centerScrollElement(pane, verses[verseIdx]);
  }

  function searchInputKeyDown(event) {
    if (event.key === "Enter") {
      btnSearchGoClick(event);
    }
  }

  function resultsBannerClick() {
    if (searchExp.state !== 'OK') {
      return;
    }
    hide(searchResults);
    show(searchDetails);
  }

  function resultsListClick(target) {
    let refSpan = target;
    if (!refSpan.matches('span.ref')) {
      refSpan = target.closest('li').querySelector('span.ref');
    }
    let decode = decodeRef(refSpan.textContent);
    setSelectedBookChapter(decode.book, decode.chapter);
    saveSelectedBookChapter();
    updatePaneRead();
    updateSelectorContents();
    highlightVerse(decode.verse);
    if (numPanes === 1) {
      hide(paneSearch);
      show(paneRead);
    }
    scrollToVerse(scrollRead, decode.verse);
  }

  function selectorBookmarksClick(event) {
    let verseIdx;
    event.preventDefault();
    let target = event.target;
    if (target === selectorBookmarks) {
      return;
    }
    if (selectingBookmarks) {
      if (activeBookmark === target) {
        deactivate(activeBookmark);
        activeBookmark = null;
        return;
      }
      if (activeBookmark) {
        deactivate(activeBookmark);
      }
      activeBookmark = target;
      activate(activeBookmark);
      return;
    }
    verseIdx = getDataVerse(target);
    goToBookmark(verseIdx);
    return;
  }

  function selectorContentsClick(event) {
    event.preventDefault();
    let target = event.target;
    if (target === selectorContents) {
      return;
    }
    target = target.closest('button');
    if (target.classList.contains('btn-book')) {
      btnBookClick(event);
      return;
    }
    if (target.classList.contains('btn-chapter')) {
      btnChapterClick(event);
      return;
    }
  }

  function selectorReadClick(event) {
    event.preventDefault();
    let target = event.target;
    if (target.classList.contains('verse-num')) {
      verseNumClick(target);
      return;
    }
    if (target.classList.contains('read-verse')) {
      readVerseClick(target);
      return;
    }
  }

  function selectorSearchClick(event) {
    event.preventDefault();
    let target = event.target;
    if (detailsBanner.contains(target)) {
      detailsBannerClick(target);
      return;
    }
    if (detailsList.contains(target)) {
      detailsListClick(target);
      return;
    }
    if (resultsBanner.contains(target)) {
      resultsBannerClick(target);
      return;
    }
    if (resultsList.contains(target)) {
      resultsListClick(target);
      return;
    }
  }

  function setActiveFontIdx() {
    let activeFont = fontClasses[activeFontIdx];
    btnFontName.textContent = fontNames[activeFontIdx];
    readVerses.classList.remove(...fontClasses);
    readVerses.classList.add(activeFont);
    resultsList.classList.remove(...fontClasses);
    resultsList.classList.add(activeFont);
    fontSample.classList.remove(...fontClasses);
    fontSample.classList.add(activeFont);
    btnFontName.classList.remove(...fontClasses);
    btnFontName.classList.add(activeFont);
  }

  function setActiveSidebarIdx(sidebarIdx) {
    activeSidebarIdx = sidebarIdx;
    saveActiveSidebarIdx();
  }

  function setSelectedBookChapter(bookIdx, chapterIdx) {
    selectedBookChapter.bookIdx = bookIdx;
    selectedBookChapter.bookName = bible.books[bookIdx][0];
    selectedBookChapter.chapterIdx = chapterIdx;
    selectedBookChapter.chapterName = (chapterIdx + 1).toString();
    selectedBookChapter.chapterCount = bible.books[bookIdx][1];
  }

  function show(element) {
    element.classList.remove('hide');
  }

  function showActiveHelp() {
    activate(helpBtns[activeHelpIdx]);
    show(helpPages[activeHelpIdx]);
  }

  function showActiveSidebar() {
    show(sidebarPanes[activeSidebarIdx]);
    if (numPanes > 1) {
      activate(sidebarBtns[activeSidebarIdx]);
    }
  }

  function sortBookmarks(sorter) {
    if (bookmarks.length === 0) {
      return;
    }
    bookmarks.sort(sorter);
    updateBookmarkList();
  }

  function updateActiveBook() {
    if (activeBook) {
      deactivate(activeBook);
    }
    activeBook = selectorContents.querySelector(
      `.btn-book[data-book="${selectedBookChapter.bookIdx}"]`);
    activate(activeBook);
  }

  function updateActiveChapter() {
    activeChapter = chapterList.querySelector(
      `.btn-chapter[data-chapter="${selectedBookChapter.chapterIdx}"]`);
    activate(activeChapter);
    activeBook.parentNode.insertBefore(chapterList, activeBook.nextSibling);
    show(chapterList);
  }

  function updateBookmarkList() {
    removeAllChildren(bookmarksList);
    if (bookmarks.length === 0) {
      show(bookmarksEmpty);
      return;
    } else {
      hide(bookmarksEmpty);
    }
    fragment = document.createDocumentFragment();
    bookmarks.forEach(function (verseIdx) {
      let btn = buildBtnBookmark(verseIdx);
      fragment.appendChild(btn);
    });
    bookmarksList.appendChild(fragment);
  }

  function updateBtnContents() {
    btnContents.textContent =
      `${selectedBookChapter.bookName} ${selectedBookChapter.chapterIdx + 1}`;
  }

  function updateBtnOt() {
    activeTestament = btnOt;
    activate(btnOt);
    deactivate(btnNt);
    show(otList);
    hide(ntList);
  }

  function updateBtnNt() {
    activeTestament = btnNt;
    deactivate(btnOt);
    activate(btnNt);
    hide(otList);
    show(ntList);
  }

  function updateBtnTestament() {
    if (selectedBookChapter.bookIdx < firstNtBookIdx) {
      updateBtnOt();
    } else {
      updateBtnNt();
    }
  }

  function updateColumns(columnCount) {
    let activeColumnBtn = columnBtns[activeColumnIdx];
    deactivate(activeColumnBtn);
    activeColumnIdx = columnCount - 1;
    saveActiveColumnIdx();
    activeColumnBtn = columnBtns[activeColumnIdx];
    activate(activeColumnBtn);
    readVerses.classList.remove('two-column', 'three-column');
    if (columnCount === 2) {
      readVerses.classList.add('two-column');
      return;
    }
    if (columnCount === 3) {
      readVerses.classList.add('three-column');
      return;
    }
  }

  function updateDetailsList() {
    removeAllChildren(detailsList);
    if (searchExp.state !== 'OK') {
      return;
    }
    let list = buildSearchDetailsList(searchExp.bins);
    detailsList.appendChild(list);
  }

  function updatePaneRead() {
    updateBtnContents();
    updateSelectorRead();
  }

  function updatePanes() {
    getNumPanes();
    if (currentPanes === numPanes) {
      return;
    }
    if (currentPanes > numPanes) {
      decreasePanes();
    } else {
      increasePanes();
    }
    currentPanes = numPanes;
  }

  function updateReadStats() {
    removeAllChildren(readStats);
    let header = document.createElement('p');
    header.classList.add('centered');
    header.textContent = 'Book / Chapter / Verse';
    readStats.appendChild(header);

    let stats = bibleStats[
      `${selectedBookChapter.bookName} ${selectedBookChapter.chapterName}`];
    let bibleCitation = document.createElement('p');
    bibleCitation.classList.add('centered', 'citation');
    bibleCitation.textContent =
      `${stats.bibleBook} / ${stats.bibleChapter} / ${stats.bibleVerse} in the Bible`;
    readStats.appendChild(bibleCitation);

    if (stats.ntBook) {
      let ntCitation = document.createElement('p');
      ntCitation.classList.add('centered', 'citation');
      ntCitation.textContent =
        `${stats.ntBook} / ${stats.ntChapter} / ${stats.ntVerse} in the New Testament`;
      readStats.appendChild(ntCitation);
    }
  }

  function updateReadVerses() {
    removeAllChildren(readVerses);
    fragment = document.createDocumentFragment();
    let chapters = bible.books[selectedBookChapter.bookIdx][4];
    let start = chapters[selectedBookChapter.chapterIdx][0];
    let end = chapters[selectedBookChapter.chapterIdx][1];
    let verseNum = 1;
    for (let idx = start; idx <= end; idx++) {
      addReadVerse(idx, verseNum);
      verseNum++;
    }
    let list = document.createElement('ul');
    list.appendChild(fragment);
    readVerses.appendChild(list);
  }

  function updateResultsBanner() {
    removeAllChildren(resultsBanner);
    if (searchExp.type === 'EMPTY') {
      resultsBanner.textContent = 'Enter a search expression.';
      return;
    }
    if (searchExp.type === 'INVALID') {
      resultsBanner.textContent = 'Invalid search expression.';
      return;
    }
    if (searchExp.bibleWords !== 'OK') {
      resultsBanner.textContent = searchExp.bibleWords;
      return;
    }
    let header = document.createElement('p');
    header.classList.add('centered');
    header.textContent = 'Occurrences / Verses';
    let citation = document.createElement('p');
    citation.classList.add('centered', 'citation');
    citation.textContent = searchFilter.citation;
    let hint = document.createElement('p');
    hint.classList.add('hint');
    hint.textContent = 'Tap here for details';
    resultsBanner.appendChild(header);
    resultsBanner.appendChild(citation);
    resultsBanner.appendChild(hint);
  }

  function updateResultsList() {
    removeAllChildren(resultsList);
    if (!searchFilter.verses) {
      return;
    }
    fragment = document.createDocumentFragment();
    searchFilter.verses.forEach(function (verse) {
      addSearchVerse(verse);
    });
    let verseList = document.createElement('ul');
    verseList.appendChild(fragment);
    resultsList.appendChild(verseList);
  }

  function updateSearchDetails() {
    updateDetailsList();
  }

  function updateSearchPane() {
    searchExp = performSearch(searchStr);
    buildBibleSearchFilter();

    updateSearchResults();
    updateSearchDetails();
    activateBibleDetail();

    show(searchResults);
    hide(searchDetails);

    scrollSearch.scrollTop = 0;
    searchInput.blur();
  }

  function updateSearchResults() {
    updateResultsBanner();
    updateResultsList();
  }

  function updateSelectorContents() {
    updateBtnContents();
    updateBtnTestament();
    updateActiveBook();
    resetChapterList(selectedBookChapter.chapterCount);
    updateActiveChapter();
  }

  function updateSelectorRead() {
    updateReadVerses();
    updateReadStats();
    show(selectorRead);
    window.scroll(0, 0);
  }

  function updateSidebar(sidebarIdx) {
    if (activeSidebarIdx !== null) {
      hideActiveSidebar();
    }
    setActiveSidebarIdx(sidebarIdx);
    showActiveSidebar();
  }

  function verseNumClick(target) {
    let verse = target.closest('li');
    let verseIdx = getDataVerse(verse);
    copyToClipboard(verseIdx);
    verse.classList.add('clipboard-flash');
    setTimeout(function () {
      verse.classList.remove('clipboard-flash');
    }, 2000);
  }

  obj.initialize = function () {
    addListeners();

    buildBibleStats();
    buildOtList();
    buildNtList();
    buildChapterList();

    getBookmarks();
    updateBookmarkList();

    getSearchStr();
    searchInput.value = searchStr;
    updateSearchPane();

    getActiveHelpIdx();
    showActiveHelp();

    getActiveFontIdx();
    setActiveFontIdx();

    getActiveColumnIdx();
    updateColumns(activeColumnIdx + 1);

    getSelectedBookChapter();
    updateSelectorContents();
    updatePaneRead();

    getNumPanes();
    getActiveSidebarIdx();
    if (numPanes > 1) {
      if (activeSidebarIdx === null) {
        setActiveSidebarIdx(sidebarContents);
      }
      updateSidebar(activeSidebarIdx);
      if (activeSidebarIdx === sidebarContents) {
        centerScrollElement(scrollContents, activeChapter);
      }
    }
  };
})(window.KJV);

KJV.initialize();
